
import java.util.Vector;

public class Board
{
    int xsize;
    int ysize;
    char [][]terrain;  // .,-,=,@  indexed by y,x
    Coordinate [][]vertices;
    Vector [][]robots;
    Vector [][]packages;
    Vector bases;

    private int trows = 0; // number of terrain rows added to board


    public Board (int xsize, int ysize)
    {
	this.xsize = xsize;
	this.ysize = ysize;

	terrain = new char[ysize][];
	vertices = new Coordinate[xsize][ysize];
	robots = new Vector[xsize][ysize];
	packages = new Vector[xsize][ysize];
	bases = new Vector();

	for (int i = 0; i < xsize; i++)
	    for (int j = 0; j < ysize; j++) {
		robots[i][j] = new Vector();
		packages[i][j] = new Vector();
	    }
		
    }


    public void addRow(String row)
    {
	int ind, pos = 0;

	while ((ind = row.indexOf("@", pos)) != -1) {
	    bases.addElement(new Base(ind, trows));
	    pos = ind + 1;
	}

	terrain[trows++] = row.toCharArray();
    }


    public void initVertices()
    {
	for (int i = 0; i < xsize; i++)
	    for (int j = 0; j < ysize; j++) {
		vertices[i][j] = new Coordinate(i, j);
		vertices[i][j].color = Coordinate.WHITE;
		vertices[i][j].distance = 0;
		vertices[i][j].parent = null;
	    }
    }


    public boolean squareIsPassable(int x, int y)
    {
	return terrain[y][x] != '#' && terrain[y][x] != '~';
    }


    public boolean squareIsWater(int x, int y)
    {
	return terrain[y][x] == '~';
    }


    public boolean enemyOnSquare(int x, int y, Robot my_bot)
    {
	Robot bot;
	boolean enemy = false;

	for (int i = 0; i < robots[x][y].size(); i++) {
	    bot = (Robot) robots[x][y].elementAt(i);
	    if (bot != my_bot) {
		enemy = true;
		break;
	    }
	}
	    
	return enemy;
    }


    public Base lookupBase(Coordinate loc)
    {
	Base base = null;
	boolean found = false;

	for (int i = 0; i < bases.size(); i++) {
	    base = (Base) bases.elementAt(i);
	    if (base.location.x == loc.x && base.location.y == loc.y) {
		found = true;
		break;
	    }
	}
	
	if (found)
	    return base;
	else
	    return null;

    }

}
