
import java.util.*;

public class Brain
{
    // AI states
    static final int GATHER_MODE = 0;
    static final int DELIVER_MODE = 1;
    
    static final int MIN_FILL_PCT = 50;


    ServerConn sc;
    Board board;
    Robot my_bot;
    int ai_state;
    int max_bot_capacity;
    int target_bot;  // TODO: Make this a Robot?
    Robot active_bot = null;
    Coordinate cur_goal = null;

    Vector bot_list = null;
    Vector package_list = null;


    public Brain(String host, int port) throws Exception
    {
	sc = new ServerConn(host, port);
	System.out.println("Connection established");
	ai_state = GATHER_MODE;
	target_bot = -1;

	bot_list = new Vector();
	package_list = new Vector();
    }


    public void playGame() throws Exception
    {
	System.out.println("Initiating game");

	sc.send("Player\n");

	// Receive and set up the game board
	recvBoard();

	// Receive and set up initial config info
	recvConfig();

	// Handle initial server response
	recvServerResponse();

	while (true) {

	    System.out.println();
	    System.out.println("Money: " + my_bot.money);

	    // Receive the list of packages in current square
	    recvPackageList();

	    // Calculate and execute next move
	    execNextMove();

	    // Set bots' alive flag to false
	    initBotsAliveFlag();

	    recvServerResponse();

	    // Remove bots that weren't sent from server
	    eliminateDeadBots();
	}
    }


    private void recvBoard() throws Exception
    {
	String line, row, padding;
	StringTokenizer tokenizer;
	int xsize, ysize;

	// Get board dimensions
	line = sc.recv();
        tokenizer = new StringTokenizer(line);

	// Pad board on all sides with impassable terrain,
	// so 2 extra rows and columns will be added

	xsize = Integer.parseInt(tokenizer.nextToken()) + 2;
	ysize = Integer.parseInt(tokenizer.nextToken()) + 2;

	board = new Board(xsize, ysize);

	padding = "#";
	for (int i = 0; i < xsize - 1; i++)
	    padding += "#";
	board.addRow(padding);

	// Get board terrain
	for (int i = 0; i < ysize - 2; i++) {
	    line = sc.recv();
	    row = "#" + line + "#";
	    board.addRow(row);
	}

	board.addRow(padding);

	System.out.println("Board received");
    }


    private void recvConfig() throws Exception
    {
	String line;
	StringTokenizer tokenizer;
	int bot_id, money;

	line = sc.recv();
	tokenizer = new StringTokenizer(line);

	bot_id = Integer.parseInt(tokenizer.nextToken());
	max_bot_capacity = Integer.parseInt(tokenizer.nextToken());
	money = Integer.parseInt(tokenizer.nextToken());

	my_bot = new Robot(bot_id);
	my_bot.money = money;
    }


    private void recvServerResponse() throws Exception
    {
	String line, tok;
	StringTokenizer tokenizer;
	int bot_id = 0, xpos, ypos;

	line = sc.recv();
	tokenizer = new StringTokenizer(line);

	tok = tokenizer.nextToken();

	while (true) {

	    if (tok.startsWith("#")) {
		bot_id = Integer.parseInt(tok.substring(1));
		active_bot = lookupBot(bot_id);

		if (active_bot == null)
		    active_bot = initBot(bot_id);

		active_bot.alive = true;

		System.out.println("Processing bot: " + active_bot.id + 
				   ", x: " + active_bot.location.x + ", y: " +
				   active_bot.location.y);

		if (!tokenizer.hasMoreTokens())
		    break;

		tok = tokenizer.nextToken().toUpperCase();

		// No command found, just the next bot #
		if (tok.startsWith("#"))
		    continue;
	    }

	    if (tok.startsWith("N") || tok.startsWith("S") ||
		tok.startsWith("E") || tok.startsWith("W")) {
		moveBot(tok);
	    }
	    else if (tok.startsWith("P")) {
		tok = tokenizer.nextToken();
		pickUp(Integer.parseInt(tok));
	    }
	    else if (tok.startsWith("D")) {
		tok = tokenizer.nextToken();
		drop(Integer.parseInt(tok));
	    }
	    else if (tok.startsWith("X")) {
		xpos = Integer.parseInt(tokenizer.nextToken());
		tok = tokenizer.nextToken();  // "Y"
		ypos = Integer.parseInt(tokenizer.nextToken());
		
		setBotPos(xpos, ypos);
	    }

	    if (!tokenizer.hasMoreTokens())
		break;

	    tok = tokenizer.nextToken().toUpperCase();
	}
    }


    private void moveBot(String dir)
    {
	board.robots[active_bot.location.x][active_bot.location.y].removeElement(active_bot);

	if (dir.startsWith("N")) {
	    active_bot.location.y++;
	}
	else if (dir.startsWith("S")) {
	    active_bot.location.y--;
	}
	else if (dir.startsWith("E")) {
	    active_bot.location.x++;
	}
	else if (dir.startsWith("W")) {
	    active_bot.location.x--;
	}

	board.robots[active_bot.location.x][active_bot.location.y].addElement(active_bot);
    }


    private void pickUp(int pack_id)
    {
	if (active_bot == my_bot) {
	    Package pack = lookupPackage(pack_id);

	    my_bot.payload.addElement(pack);
	    my_bot.carried_wt += pack.weight;
	}
    }


    private void drop(int pack_id)
    {
	if (active_bot == my_bot) {
	    Package pack = lookupPayloadPackage(pack_id);

	    if (active_bot == my_bot)
		my_bot.carried_wt -= pack.weight;

	    active_bot.payload.removeElement(pack);
	}
    }


    private Robot initBot(int bot_id)
    {
	Robot bot;

	if (bot_id != my_bot.id)
	    bot = new Robot(bot_id);
	else
	    bot = my_bot;
	
	bot_list.addElement(bot);

	return bot;
    }


    private void setBotPos(int xpos, int ypos)
    {
	active_bot.setLocation(xpos, ypos);
    }


    private Robot lookupBot(int bot_id)
    {
	Robot bot = null;
	boolean found = false;

	if (bot_id == my_bot.id) {
	    bot = my_bot;
	    found = true;
	}
	else {
	    for (int i = 0; i < bot_list.size(); i++) {
		bot = (Robot) bot_list.elementAt(i);
		if (bot.id == bot_id) {
		    found = true;
		    break;
		}
	    }
	}
	
	if (found)
	    return bot;
	else
	    return null;
    }


    private Package lookupPackage(int pack_id)
    {
	Package pack = null;
	boolean found = false;

	for (int i = 0; i < package_list.size(); i++) {
	    pack = (Package) package_list.elementAt(i);
	    if (pack.id == pack_id) {
		found = true;
		break;
	    }
	}
	
	if (found)
	    return pack;
	else
	    return null;
    }


    private Package lookupPayloadPackage(int pack_id)
    {
	Package pack = null;
	boolean found = false;

	for (int i = 0; i < my_bot.payload.size(); i++) {
	    pack = (Package) my_bot.payload.elementAt(i);
	    if (pack.id == pack_id) {
		found = true;
		break;
	    }
	}
	
	if (found)
	    return pack;
	else
	    return null;
    }


    private void recvPackageList() throws Exception
    {
	String line;
	StringTokenizer tokenizer;
	int pack_id, xdest, ydest, wt;
	Package pack;

	package_list = new Vector();

	line = sc.recv();
	tokenizer = new StringTokenizer(line);

	while (tokenizer.hasMoreTokens()) {

	    pack_id = Integer.parseInt(tokenizer.nextToken());
	    xdest = Integer.parseInt(tokenizer.nextToken());
	    ydest = Integer.parseInt(tokenizer.nextToken());
	    wt = Integer.parseInt(tokenizer.nextToken());	    

	    pack = new Package(pack_id);
	    package_list.addElement(pack);
	    pack.dest = new Coordinate(xdest, ydest);
	    pack.weight = wt;
	}

	System.out.println("Package list received");
    }


    private void execNextMove() throws Exception
    {
	String move = null;
	boolean has_loot = false;
	Base base;

	if (my_bot.carried_wt > 0)
	    ai_state = DELIVER_MODE;
	else
	    ai_state = GATHER_MODE;

	// Check for any easy kills
	if ((move = getKillCmd()) != null) {
	    System.out.println("Attempting kill");
	}
	
	// Check for any package dropoffs
	if (move == null)
	    move = getDropCmd();

	if (move == null && cur_goal != null && 
	    my_bot.location.x == cur_goal.x &&
	    my_bot.location.y == cur_goal.y) {

	    if (ai_state == GATHER_MODE) {
		move = getLootCmd();

		// Invalidate base if empty
		if (move == null && ((base = board.lookupBase(my_bot.location)) != null)) {
		    System.out.println("Setting base unusable: " + base.location.x + 
				       ", " + base.location.y);
		    base.usable = false;
		}
	    }
	}

	// Check for any packages strewn about
	if (move == null)
	    move = getLootCmd();
	
	if (move == null) {
	    cur_goal = acquireGoal();
	    plotPath(my_bot.location, cur_goal);
	    move = getMoveCmd();
	}

	sc.send(move);
    }


    private String getLootCmd()
    {
	Package pack;
	boolean retval = false;
	String loot = null, move = null;
	int rem_wt;

	rem_wt = max_bot_capacity - my_bot.carried_wt;

	for (int i = 0; i < package_list.size(); i++) {
	    pack = (Package) package_list.elementAt(i);
	    if (pack.weight <= rem_wt) {

		System.out.println("Picking up " + pack.id + ", Wt: " + pack.weight +
				   ", Max: " + max_bot_capacity);

		if (loot == null)
		    loot = pack.id + " ";
		else
		    loot += pack.id + " ";

		rem_wt -= pack.weight;
	    }
	}

	if (loot != null)
	    move = "1 Pick " + loot.substring(0, loot.length() - 1) + "\n";

	if (move != null)
	    my_bot.money--;

	return move;
    }


    private String getDropCmd()
    {
	Package pack = null;
	String loot = null, move = null;

	for (int i = 0; i < my_bot.payload.size(); i++) {

	    pack = (Package) my_bot.payload.elementAt(i);

	    if (my_bot.location.x == pack.dest.x &&
		my_bot.location.y == pack.dest.y) {

		System.out.println("Dropping " + pack.id + ", Wt: " + pack.weight);

		if (loot == null)
		    loot = pack.id + " ";
		else
		    loot += pack.id + " ";
	    }
	}

	if (loot != null)
	    move = "1 Drop " + loot.substring(0, loot.length() - 1) + "\n";

	if (move != null)
	    my_bot.money--;

	return move;
    }


    private String getKillCmd()
    {
	String move = null;
	Coordinate coord = my_bot.location;

	if (board.enemyOnSquare(coord.x, coord.y + 1, my_bot) &&
	    board.squareIsWater(coord.x, coord.y + 2))
	    move = "2 Move N\n";

	else if (board.enemyOnSquare(coord.x, coord.y - 1, my_bot) &&
		 board.squareIsWater(coord.x, coord.y - 2))	    
	    move = "2 Move S\n";

	else if (board.enemyOnSquare(coord.x + 1, coord.y, my_bot) &&
		 board.squareIsWater(coord.x + 2, coord.y))	    
	    move = "2 Move E\n";

	else if (board.enemyOnSquare(coord.x - 1, coord.y, my_bot) &&
		 board.squareIsWater(coord.x - 2, coord.y))	    
	    move = "2 Move W\n";

	if (move != null)
	    my_bot.money -= 2;

	return move;
    }


    private Coordinate acquireGoal()
    {
	Base b;
	Coordinate goal = null;
	Package pack = null;
	int x, y, rand;
	boolean goal_ok;
	Random r;


	if (ai_state == GATHER_MODE) {
	    for (int i = 0; i < board.bases.size(); i++) {
		b = (Base) board.bases.elementAt(i);
		if (b.usable) {
		    goal = b.location;
		    break;
		}
	    }
	}
	else if (ai_state == DELIVER_MODE) {
	    pack = (Package) my_bot.payload.elementAt(0);
	    goal = pack.dest;
	}

	if (goal == null) {
	    
	    System.out.println("Acquiring random goal");

	    r = new Random();

	    while (true) {

		x = my_bot.location.x;
		y = my_bot.location.y;

		rand = (int) (r.nextFloat() * 3.0);

		switch (rand) {
		case 0: // N
		    y++;
		    break;
		case 1:  // S
		    y--;
		    break;
		case 2:  // E
		    x++;
		    break;
		case 3:  // W
		    x--;
		    break;		    
		}
		
		if (board.squareIsPassable(x, y))
		    break;
	    }
	    goal = new Coordinate(x, y);
	}

	System.out.println("Acquired goal: " + goal.x + ", " + goal.y);

	return goal;
    }


    private void plotPath(Coordinate from, Coordinate to)
    {
	Vector visited = new Vector();
	Coordinate check_coord, v;

	board.initVertices();	
	board.vertices[from.x][from.y].color = Coordinate.GRAY;
	visited.addElement(board.vertices[from.x][from.y]);

	while (visited.size() != 0) {
	    check_coord = (Coordinate) visited.elementAt(0);
	    visited.removeElementAt(0);

	    if (check_coord.x == to.x && check_coord.y == to.y)  // found
		break;

	    // N
	    if (board.squareIsPassable(check_coord.x, check_coord.y + 1)) {
		v = board.vertices[check_coord.x][check_coord.y + 1];
		if (v.color == Coordinate.WHITE) {
		    v.color = Coordinate.GRAY;
		    v.distance = check_coord.distance + 1;
		    v.parent = check_coord;
		    visited.addElement(v);
		}
	    }
	    // S
	    if (board.squareIsPassable(check_coord.x, check_coord.y - 1)) {
		v = board.vertices[check_coord.x][check_coord.y - 1];
		if (v.color == Coordinate.WHITE) {
		    v.color = Coordinate.GRAY;
		    v.distance = check_coord.distance + 1;
		    v.parent = check_coord;
		    visited.addElement(v);
		}
	    }
	    // E
	    if (board.squareIsPassable(check_coord.x + 1, check_coord.y)) {
		v = board.vertices[check_coord.x + 1][check_coord.y];
		if (v.color == Coordinate.WHITE) {
		    v.color = Coordinate.GRAY;
		    v.distance = check_coord.distance + 1;
		    v.parent = check_coord;
		    visited.addElement(v);
		}
	    }
	    // W
	    if (board.squareIsPassable(check_coord.x - 1, check_coord.y)) {
		v = board.vertices[check_coord.x - 1][check_coord.y];
		if (v.color == Coordinate.WHITE) {
		    v.color = Coordinate.GRAY;
		    v.distance = check_coord.distance + 1;
		    v.parent = check_coord;
		    visited.addElement(v);
		}
	    }
	    
	    check_coord.color = Coordinate.BLACK;
	}

    }


    private String getMoveCmd()
    {
	Coordinate coord, last_coord = null;
	String move, dir = null;
	int bid = 1;

	System.out.println("Distance to goal: " + 
			   board.vertices[cur_goal.x][cur_goal.y].distance);

	coord = board.vertices[cur_goal.x][cur_goal.y];

	while (coord.parent != null) {
	    last_coord = coord;
	    coord = coord.parent;
	}

	System.out.println("my_bot: " + my_bot.location.x + ", " + my_bot.location.y);
	System.out.println("first path step: " + last_coord.x + ", " + last_coord.y);

	if (killDanger(my_bot.location)) {
	    System.out.println("Detected possible kill danger, increasing bid");
	    bid = 3;
	}
	if (killDanger(last_coord)) {
	    System.out.println("Detected possible kill danger, skipping move");
	    move = bid + " Drop\n"; // do nothing
	}
	else {

	    if (last_coord.y > my_bot.location.y)
		dir = "N";
	    else if (last_coord.y < my_bot.location.y)
		dir = "S";
	    else if (last_coord.x > my_bot.location.x)
		dir = "E";
	    else if (last_coord.x < my_bot.location.x)
		dir = "W";

	    move = bid + " Move " + dir + "\n";
	}

	if (move != null)
	    my_bot.money -= bid;

	return move;
    }


    private boolean killDanger(Coordinate coord)
    {
	boolean danger = false;

	if (board.squareIsWater(coord.x, coord.y + 1) &&
	    board.enemyOnSquare(coord.x, coord.y - 1, my_bot))
	    danger = true;
	else if (board.squareIsWater(coord.x, coord.y - 1) &&
		 board.enemyOnSquare(coord.x, coord.y + 1, my_bot))
	    danger = true;
	else if (board.squareIsWater(coord.x + 1, coord.y) &&
		 board.enemyOnSquare(coord.x - 1, coord.y, my_bot))
	    danger = true;
	else if (board.squareIsWater(coord.x - 1, coord.y) &&
		 board.enemyOnSquare(coord.x + 1, coord.y, my_bot))
	    danger = true;

	return danger;
    }


    private void initBotsAliveFlag()
    {
	Robot bot;

	for (int i = 0; i < bot_list.size(); i++) {
	    bot = (Robot) bot_list.elementAt(i);
	    bot.alive = false;
	}
    }


    private void eliminateDeadBots()
    {
	Robot bot;

	for (int i = 0; i < bot_list.size(); i++) {
	    bot = (Robot) bot_list.elementAt(i);
	    if (bot.alive == false) {
		bot_list.removeElement(bot);
		board.robots[bot.location.x][bot.location.y].removeElement(bot);
	    }
	}		
	
    }

}
